##############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##############################################################################

import concurrent.futures
import datetime
import io
import math
import pickle
import time

import pymongo
from girder_large_image.models.image_item import ImageItem

import large_image
from girder import logger
from girder.constants import AccessType, SortDir
from girder.models.file import File
from girder.models.item import Item
from girder.models.model_base import Model
from girder.models.upload import Upload

# Some annotation elements can be very large.  If they pass a size threshold,
# store part of them in an associated file.  This is slower, so don't do it for
# small ones.
MAX_ELEMENT_CHECK = 100
MAX_ELEMENT_DOCUMENT = 10000
MAX_ELEMENT_USER_DOCUMENT = 1000000


class Annotationelement(Model):
    bboxKeys = {
        'left': ('bbox.highx', '$gte'),
        'right': ('bbox.lowx', '$lt'),
        'top': ('bbox.highy', '$gte'),
        'bottom': ('bbox.lowy', '$lt'),
        'low': ('bbox.highz', '$gte'),
        'high': ('bbox.lowz', '$lt'),
        'minimumSize': ('bbox.size', '$gte'),
        'size': ('bbox.size', None),
        'details': ('bbox.details', None),
    }

    def initialize(self):
        self.name = 'annotationelement'
        self.ensureIndices([
            'annotationId',
            '_version',
            ([
                ('annotationId', SortDir.ASCENDING),
                ('bbox.lowx', SortDir.DESCENDING),
                ('bbox.highx', SortDir.ASCENDING),
                ('bbox.size', SortDir.DESCENDING),
            ], {
                'name': 'annotationBboxIdx',
            }),
            ([
                ('annotationId', SortDir.ASCENDING),
                ('bbox.size', SortDir.DESCENDING),
            ], {
                'name': 'annotationBboxSizeIdx',
            }),
            ([
                ('annotationId', SortDir.ASCENDING),
                ('_version', SortDir.DESCENDING),
                ('element.group', SortDir.ASCENDING),
            ], {
                'name': 'annotationGroupIdx',
            }),
            ([
                ('created', SortDir.ASCENDING),
                ('_version', SortDir.ASCENDING),
            ], {}),
            'element.girderId',
        ])

        self.exposeFields(AccessType.READ, (
            '_id', '_version', 'annotationId', 'created', 'element'))
        self.versionId = None

    def getNextVersionValue(self):
        """
        Maintain a version number.  This is a single sequence that can be used
        to ensure we have the correct set of elements for an annotation.

        :returns: an integer version number that is strictly increasing.
        """
        version = None
        if self.versionId is not None:
            version = self.collection.find_one_and_update(
                {'_id': self.versionId},
                {'$inc': {'_version': 1}})
        if version is None:
            versionObject = self.collection.find_one(
                {'annotationId': 'version_sequence'})
            if versionObject is None:
                startingId = self.collection.find_one({}, sort=[('_version', SortDir.DESCENDING)])
                startingId = startingId['_version'] + 1 if startingId else 0
                self.versionId = self.collection.insert_one(
                    {'annotationId': 'version_sequence', '_version': startingId},
                ).inserted_id
            else:
                self.versionId = versionObject['_id']
            version = self.collection.find_one_and_update(
                {'_id': self.versionId},
                {'$inc': {'_version': 1}})
        return version['_version']

    def getElements(self, annotation, region=None):
        """
        Given an annotation, fetch the elements from the database and add them
        to it.

        When a region is used to request specific element, the following
        keys can be specified:

            :left, right, top, bottom, low, high: the spatial area where
                elements are located, all in pixels.  If an element's bounding
                box is at least partially within the requested area, that
                element is included.
            :minimumSize: the minimum size of an element to return.
            :sort, sortdir: standard sort options.  The sort key can include
                size and details.
            :limit: limit the total number of elements by this value.  Defaults
                to no limit.
            :offset: the offset within the query to start returning values.  If
                maxDetails is used, to get subsequent sets of elements, the
                offset needs to be increased by the actual number of elements
                returned from a previous query, which will vary based on the
                details of the elements.
            :maxDetails: if specified, limit the total number of elements by
                the sum of their details values.  This is applied in addition
                to limit.  The sum of the details values of the elements may
                exceed maxDetails slightly (the sum of all but the last element
                will be less than maxDetails, but the last element may exceed
                the value).
            :minElements: if maxDetails is specified, always return this many
                elements even if they are very detailed.
            :centroids: if specified and true, only return the id, center of
                the bounding box, and bounding box size for each element.

        :param annotation: the annotation to get elements for.  Modified.
        :param region: if present, a dictionary restricting which annotations
            are returned.
        """
        annotation['_elementQuery'] = {}
        annotation['annotation']['elements'] = list(self.yieldElements(
            annotation, region, annotation['_elementQuery']))

    def yieldElements(self, annotation, region=None, info=None, bbox=False):  # noqa
        """
        Given an annotation, fetch the elements from the database.

        When a region is used to request specific element, the following
        keys can be specified:

            :left, right, top, bottom, low, high: the spatial area where
                elements are located, all in pixels.  If an element's bounding
                box is at least partially within the requested area, that
                element is included.
            :minimumSize: the minimum size of an element to return.
            :sort, sortdir: standard sort options.  The sort key can include
                size and details.
            :limit: limit the total number of elements by this value.  Defaults
                to no limit.
            :offset: the offset within the query to start returning values.  If
                maxDetails is used, to get subsequent sets of elements, the
                offset needs to be increased by the actual number of elements
                returned from a previous query, which will vary based on the
                details of the elements.
            :maxDetails: if specified, limit the total number of elements by
                the sum of their details values.  This is applied in addition
                to limit.  The sum of the details values of the elements may
                exceed maxDetails slightly (the sum of all but the last element
                will be less than maxDetails, but the last element may exceed
                the value).
            :minElements: if maxDetails is specified, always return this many
                elements even if they are very detailed.
            :centroids: if specified and true, only return the id, center of
                the bounding box, and bounding box size for each element.
            :bbox: if specified and true and centroids are not specified,
                add _bbox to each element with the bounding box record.

        :param annotation: the annotation to get elements for.  Modified.
        :param region: if present, a dictionary restricting which annotations
            are returned.
        :param info: an optional dictionary that will be modified with
            additional query information, including count (total number of
            available elements), returned (number of elements in response),
            maxDetails (as specified by the region dictionary), details (sum of
            details returned), limit (as specified by region), centroids (a
            boolean based on the region specification).
        :param bbox: if True, always return bounding box information.
        :returns: a list of elements.  If centroids were requested, each entry
            is a list with str(id), x, y, size.  Otherwise, each entry is the
            element record.
        """
        info = info if info is not None else {}
        region = region or {}
        query = {
            'annotationId': annotation.get('_annotationId', annotation['_id']),
            '_version': annotation['_version'],
        }
        for key in region:
            if key in self.bboxKeys and self.bboxKeys[key][1]:
                if self.bboxKeys[key][1] == '$gte' and float(region[key]) <= 0:
                    continue
                query[self.bboxKeys[key][0]] = {
                    self.bboxKeys[key][1]: float(region[key])}
        if region.get('sort') in self.bboxKeys:
            sortkey = self.bboxKeys[region['sort']][0]
        else:
            sortkey = region.get('sort') or '_id'
        sortdir = int(region['sortdir']) if region.get('sortdir') else SortDir.ASCENDING
        limit = int(region['limit']) if region.get('limit') else 0
        maxDetails = int(region.get('maxDetails') or 0)
        minElements = int(region.get('minElements') or 0)
        queryLimit = max(minElements, maxDetails) if maxDetails and (
            not limit or max(minElements, maxDetails) < limit) else limit
        offset = int(region['offset']) if region.get('offset') else 0
        logger.debug('element query %r for %r', query, region)
        fields = {'_id': True, 'element': True, 'bbox.details': True, 'datafile': True}
        centroids = str(region.get('centroids')).lower() == 'true'
        if centroids:
            # fields = {'_id': True, 'element': True, 'bbox': True}
            fields = {
                '_id': True,
                'element.id': True,
                'bbox': True}
            proplist = []
            propskeys = ['type', 'fillColor', 'lineColor', 'lineWidth', 'closed']
            # This should match the javascript
            defaultProps = {
                'fillColor': 'rgba(0,0,0,0)',
                'lineColor': 'rgb(0,0,0)',
                'lineWidth': 2,
            }
            for key in propskeys:
                fields['element.%s' % key] = True
            props = {}
            info['centroids'] = True
            info['props'] = proplist
            info['propskeys'] = propskeys
        elif region.get('bbox'):
            fields.pop('bbox.details')
            fields['bbox'] = True
        if bbox:
            fields.pop('bbox.details', None)
            fields['bbox'] = True
        elementCursor = self.find(
            query=query, sort=[(sortkey, sortdir)], limit=queryLimit,
            offset=offset, fields=fields)

        info.update({
            'count': elementCursor.count(),
            'offset': offset,
            'filter': query,
            'sort': [sortkey, sortdir],
        })
        details = count = 0
        if maxDetails:
            info['maxDetails'] = maxDetails
        if minElements:
            info['minElements'] = minElements
        if limit:
            info['limit'] = limit
        for entry in elementCursor:
            element = entry['element']
            element.setdefault('id', entry['_id'])
            if centroids:
                bbox = entry.get('bbox')
                if not bbox or 'lowx' not in bbox or 'size' not in bbox:
                    continue
                prop = tuple(
                    element.get(key, defaultProps.get(key)) for key in propskeys
                    if element.get(key, defaultProps.get(key)) is not None)
                if prop not in props:
                    props[prop] = len(props)
                    proplist.append(list(prop))
                yield [
                    str(element['id']),
                    (bbox['lowx'] + bbox['highx']) / 2,
                    (bbox['lowy'] + bbox['highy']) / 2,
                    bbox['size'] if entry.get('type') != 'point' else 0,
                    props[prop],
                ]
                details += 1
            else:
                if entry.get('datafile'):
                    datafile = entry['datafile']
                    data = io.BytesIO()
                    chunksize = 1024 ** 2
                    with File().open(File().load(datafile['fileId'], force=True)) as fptr:
                        while True:
                            chunk = fptr.read(chunksize)
                            if not len(chunk):
                                break
                            data.write(chunk)
                    data.seek(0)
                    element[datafile['key']] = pickle.load(data)
                    if 'userFileId' in datafile:
                        data = io.BytesIO()
                        chunksize = 1024 ** 2
                        with File().open(File().load(datafile['userFileId'], force=True)) as fptr:
                            while True:
                                chunk = fptr.read(chunksize)
                                if not len(chunk):
                                    break
                                data.write(chunk)
                        data.seek(0)
                        element['user'] = pickle.load(data)
                if region.get('bbox') and 'bbox' in entry:
                    element['_bbox'] = entry['bbox']
                    if 'bbox' not in info:
                        info['bbox'] = {}
                    for axis in {'x', 'y', 'z'}:
                        lkey, hkey = 'low' + axis, 'high' + axis
                        if lkey in entry['bbox'] and hkey in entry['bbox']:
                            info['bbox'][lkey] = min(
                                info['bbox'].get(lkey, entry['bbox'][lkey]), entry['bbox'][lkey])
                            info['bbox'][hkey] = max(
                                info['bbox'].get(hkey, entry['bbox'][hkey]), entry['bbox'][hkey])
                elif bbox and 'bbox' in entry:
                    element['_bbox'] = entry['bbox']
                yield element
                details += entry.get('bbox', {}).get('details', 1)
            count += 1
            if maxDetails and details >= maxDetails and count >= minElements:
                break
        info['returned'] = count
        info['details'] = details

    def removeWithQuery(self, query):
        """
        Remove all documents matching a given query from the collection.
        For safety reasons, you may not pass an empty query.

        Note: this does NOT return a Mongo DeleteResult.

        :param query: The search query for documents to delete,
            see general MongoDB docs for "find()"
        :type query: dict
        """
        if not query:
            msg = 'query must be specified'
            raise Exception(msg)

        attachedQuery = query.copy()
        attachedQuery['datafile'] = {'$exists': True}
        for element in self.collection.find(attachedQuery):
            for key in {'fileId', 'userFileId'}:
                if key in element['datafile']:
                    file = File().load(element['datafile'][key], force=True)
                    if file:
                        File().remove(file)
        self.collection.bulk_write([pymongo.DeleteMany(query)], ordered=False)

    def removeElements(self, annotation):
        """
        Remove all elements related to the specified annotation.

        :param annotation: the annotation to remove elements from.
        """
        self.removeWithQuery({'annotationId': annotation['_id']})

    def removeOldElements(self, annotation, oldversion=None):
        """
        Remove all elements related to the specified annotation.

        :param annotation: the annotation to remove elements from.
        :param oldversion: if present, remove versions up to this number.  If
                           none, remove versions earlier than the version in
                           the annotation record.
        """
        query = {'annotationId': annotation['_id']}
        if oldversion is None or oldversion >= annotation['_version']:
            query['_version'] = {'$lt': annotation['_version']}
        else:
            query['_version'] = {'$lte': oldversion}
        self.removeWithQuery(query)

    def _overlayBounds(self, overlayElement):
        """
        Compute bounding box information in the X-Y plane for an
        image overlay element.

        This uses numpy to perform the specified transform on the given girder
        image item in order to obtain bounding box coordinates.

        :param overlayElement: An annotation element of type 'image'.
        :returns: a tuple with 4 values: lowx, highx, lowy, highy. Runtime exceptions
         during loading the image metadata will result in the tuple (0, 0, 0, 0).
        """
        if overlayElement.get('type') not in ['image', 'pixelmap']:
            msg = ('Function _overlayBounds only accepts annotation elements '
                   'of type "image", "pixelmap."')
            raise ValueError(msg)

        import numpy as np
        lowx = highx = lowy = highy = 0

        try:
            overlayItemId = overlayElement.get('girderId')
            imageItem = ImageItem().load(overlayItemId, force=True)
            overlayImageMetadata = ImageItem().getMetadata(imageItem)
            corners = [
                [0, 0],
                [0, overlayImageMetadata['sizeY']],
                [overlayImageMetadata['sizeX'], overlayImageMetadata['sizeY']],
                [overlayImageMetadata['sizeX'], 0],
            ]
            transform = overlayElement.get('transform', {})
            transformMatrix = np.array(transform.get('matrix', [[1, 0], [0, 1]]))
            corners = [np.matmul(np.array(corner), transformMatrix) for corner in corners]
            offsetArray = np.array([transform.get('xoffset', 0), transform.get('yoffset', 0)])
            corners = [np.add(corner, offsetArray) for corner in corners]
            # use .item() to convert back to native python types
            lowx = min([corner[0] for corner in corners]).item()
            highx = max([corner[0] for corner in corners]).item()
            lowy = min([corner[1] for corner in corners]).item()
            highy = max([corner[1] for corner in corners]).item()
        except Exception:
            logger.exception('Error generating bounding box for image overlay annotation')
        return lowx, highx, lowy, highy

    def _boundingBox(self, element):
        """
        Compute bounding box information for an annotation element.

        This computes the enclosing bounding box of an element.  For points, an
        small non-zero-area region is used centered on the point.
        Additionally, a metric is stored for the complexity of the element.
        The size of the bounding box's x-y diagonal is also stored.

        :param element: the element to compute the bounding box for.
        :returns: the bounding box dictionary.  This contains 'lowx', 'lowy',
            'lowz', 'highx', 'highy', and 'highz, which are the minimum and
            maximum values in each dimension, 'details' with the complexity of
            the element, and 'size' with the x-y diagonal size of the bounding
            box.
        """
        bbox = {}
        if 'points' in element:
            pts = element['points']
            p0 = [p[0] for p in pts]
            p1 = [p[1] for p in pts]
            p2 = [p[2] for p in pts]
            bbox['lowx'] = min(p0)
            bbox['lowy'] = min(p1)
            bbox['lowz'] = min(p2)
            bbox['highx'] = max(p0)
            bbox['highy'] = max(p1)
            bbox['highz'] = max(p2)
            bbox['details'] = len(pts)
        elif element.get('type') == 'griddata':
            x0, y0, z = element['origin']
            isElements = element.get('interpretation') == 'choropleth'
            x1 = x0 + element['dx'] * (element['gridWidth'] - (1 if not isElements else 0))
            y1 = y0 + element['dy'] * (math.ceil(len(element['values']) / element['gridWidth']) -
                                       (1 if not isElements else 0))
            bbox['lowx'] = min(x0, x1)
            bbox['lowy'] = min(y0, y1)
            bbox['lowz'] = bbox['highz'] = z
            bbox['highx'] = max(x0, x1)
            bbox['highy'] = max(y0, y1)
            bbox['details'] = len(element['values'])
        elif element.get('type') in ['image', 'pixelmap']:
            lowx, highx, lowy, highy = Annotationelement()._overlayBounds(element)
            bbox['lowz'] = bbox['highz'] = 0
            bbox['lowx'] = lowx
            bbox['highx'] = highx
            bbox['lowy'] = lowy
            bbox['highy'] = highy
            bbox['details'] = 1
        else:
            center = element['center']
            bbox['lowz'] = bbox['highz'] = center[2]
            if 'width' in element:
                w = element['width'] * 0.5
                h = element['height'] * 0.5
                if element.get('rotation'):
                    absin = abs(math.sin(element['rotation']))
                    abcos = abs(math.cos(element['rotation']))
                    w, h = max(abcos * w, absin * h), max(absin * w, abcos * h)
                bbox['lowx'] = center[0] - w
                bbox['lowy'] = center[1] - h
                bbox['highx'] = center[0] + w
                bbox['highy'] = center[1] + h
                bbox['details'] = 4
            elif 'radius' in element:
                rad = element['radius']
                bbox['lowx'] = center[0] - rad
                bbox['lowy'] = center[1] - rad
                bbox['highx'] = center[0] + rad
                bbox['highy'] = center[1] + rad
                bbox['details'] = 4
            else:
                # This is a fall back for points.  Although they have no
                # dimension, make the bounding box have some extent.
                bbox['lowx'] = center[0] - 0.5
                bbox['lowy'] = center[1] - 0.5
                bbox['highx'] = center[0] + 0.5
                bbox['highy'] = center[1] + 0.5
                bbox['details'] = 1
        bbox['size'] = (
            (bbox['highy'] - bbox['lowy'])**2 +
            (bbox['highx'] - bbox['lowx'])**2) ** 0.5
        # we may want to store perimeter or area as that could help when we
        # simplify to points
        return bbox

    def _entryIsLarge(self, entry):
        """
        Return True is an entry is alrge enough it might not fit in a mongo
        document.

        :param entry: the entry to check.
        :returns: True if the entry is large.
        """
        if len(entry['element'].get('points', entry['element'].get(
                'values', []))) > MAX_ELEMENT_DOCUMENT:
            return True
        if ('user' in entry['element'] and
                len(pickle.dumps(entry['element'], protocol=4)) > MAX_ELEMENT_USER_DOCUMENT):
            return True
        return False

    def saveElementAsFile(self, annotation, entries):
        """
        If an element has a large points or values array, save that array to an
        attached file.

        :param annotation: the parent annotation.
        :param entries: the database entries document.  Modified.
        """
        item = Item().load(annotation['itemId'], force=True)
        for idx, entry in enumerate(entries[:MAX_ELEMENT_CHECK]):
            if not self._entryIsLarge(entry):
                continue
            element = entry['element'].copy()
            entries[idx]['element'] = element
            key = 'points' if 'points' in element else 'values'
            # Use the highest protocol support by all python versions we
            # support
            data = pickle.dumps(element.pop(key), protocol=4)
            elementFile = Upload().uploadFromFile(
                io.BytesIO(data), size=len(data), name='_annotationElementData',
                parentType='item', parent=item, user=None,
                mimeType='application/json', attachParent=True)
            userdata = None
            if 'user' in element:
                userdata = pickle.dumps(element.pop('user'), protocol=4)
                userFile = Upload().uploadFromFile(
                    io.BytesIO(userdata), size=len(userdata), name='_annotationElementUserData',
                    parentType='item', parent=item, user=None,
                    mimeType='application/json', attachParent=True)
            entry['datafile'] = {
                'key': key,
                'fileId': elementFile['_id'],
            }
            if userdata:
                entry['datafile']['userFileId'] = userFile['_id']
            logger.debug('Storing element as file (%r)', entry)

    def updateElementChunk(self, elements, chunk, chunkSize, annotation, now):
        """
        Update the database for a chunk of elements.  See the updateElements
        method for details.
        """
        lastTime = time.time()
        chunkStartTime = time.time()
        entries = [{
            'annotationId': annotation['_id'],
            '_version': annotation['_version'],
            'created': now,
            'bbox': self._boundingBox(element),
            'element': element,
        } for element in elements[chunk:chunk + chunkSize]]
        prepTime = time.time() - chunkStartTime
        if (len(entries) <= MAX_ELEMENT_CHECK and any(
                self._entryIsLarge(entry) for entry in entries[:MAX_ELEMENT_CHECK])):
            self.saveElementAsFile(annotation, entries)
        res = self.collection.insert_many(entries, ordered=False)
        for pos, entry in enumerate(entries):
            if 'id' not in entry['element']:
                entry['element']['id'] = str(res.inserted_ids[pos])
        # If the insert is slow, log information about it.
        if time.time() - lastTime > 10:
            logger.info('insert %d elements in %4.2fs (prep time %4.2fs), chunk %d/%d' % (
                len(entries), time.time() - chunkStartTime, prepTime,
                chunk + len(entries), len(elements)))
            lastTime = time.time()

    def updateElements(self, annotation):
        """
        Given an annotation, extract the elements from it and update the
        database of them.

        :param annotation: the annotation to save elements for.  Modified.
        """
        startTime = time.time()
        elements = annotation['annotation'].get('elements', [])
        if not len(elements):
            return
        now = datetime.datetime.now(datetime.timezone.utc)
        threads = large_image.config.cpu_count()
        chunkSize = int(max(100000 // threads, 10000))
        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as pool:
            for chunk in range(0, len(elements), chunkSize):
                pool.submit(self.updateElementChunk, elements, chunk, chunkSize, annotation, now)
        if time.time() - startTime > 10:
            logger.info('inserted %d elements in %4.2fs' % (
                len(elements), time.time() - startTime))

    def getElementGroupSet(self, annotation):
        query = {
            'annotationId': annotation.get('_annotationId', annotation['_id']),
            '_version': annotation['_version'],
        }
        groups = sorted([
            group for group in self.collection.distinct('element.group', filter=query)
            if isinstance(group, str)
        ])
        query['element.group'] = None
        if self.collection.find_one(query):
            groups.append(None)
        return groups



import itertools
import json
import math
import re

from bson.objectid import ObjectId

from girder import logger
from girder.constants import AccessType
from girder.models.folder import Folder


class AnnotationGeoJSON:
    """
    Generate GeoJSON for an annotation via an iterator.
    """

    def __init__(self, annotationId, asFeatures=False, mustConvert=False):
        """
        Return an itertor for converting an annotation into geojson.

        :param annotatioId: the id of the annotation.  No permissions checks
            are performed.
        :param asFeatures: if False, return a geojson string.  If True, return
            the features of the geojson.  This can be wrapped in
            `{'type': 'FeatureCollection', 'features: [...output...]}`
            to make it a full geojson object.
        :param mustConvert: if True, raise an exception if any annotation
            elements cannot be converted.  Otherwise, skip those elements.
        """
        from ..models.annotation import Annotation
        from ..models.annotationelement import Annotationelement

        self._id = annotationId
        self.annotation = Annotation().load(id=self._id, force=True, getElements=False)
        self.elemIterator = Annotationelement().yieldElements(self.annotation)
        self.stage = 'header'
        self.first = self.annotation['annotation']
        self.asFeatures = asFeatures
        self.mustConvert = mustConvert

    def __iter__(self):
        from ..models.annotationelement import Annotationelement

        self.elemIterator = Annotationelement().yieldElements(self.annotation)
        self.stage = 'header'
        return self

    def __next__(self):
        if self.stage == 'header':
            self.stage = 'firstelement'
            if not self.asFeatures:
                return '{"type":"FeatureCollection","features":['
        if self.stage == 'done':
            raise StopIteration
        try:
            while True:
                element = next(self.elemIterator)
                result = self.elementToGeoJSON(element)
                if result is not None:
                    break
                if self.mustConvert:
                    msg = f'Element of type {element["type"]} cannot be represented as geojson'
                    raise Exception(msg)
            prefix = ''
            if self.stage == 'firstelement':
                result['properties']['annotation'] = self.first
                self.stage = 'elements'
            else:
                prefix = ','
            if not self.asFeatures:
                return prefix + json.dumps(result, separators=(',', ':'))
            return result
        except StopIteration:
            self.stage = 'done'
            if not self.asFeatures:
                return ']}'
            raise

    def rotate(self, r, cx, cy, x, y, z):
        if not r:
            return [x, y, z]
        cosr = math.cos(r)
        sinr = math.sin(r)
        x -= cx
        y -= cy
        return [x * cosr - y * sinr + cx, x * sinr + y * sinr + cy, z]

    def circleType(self, element, geom, prop):
        x, y, z = element['center']
        r = element['radius']
        geom['type'] = 'Polygon'
        geom['coordinates'] = [[
            [x - r, y - r, z],
            [x + r, y - r, z],
            [x + r, y + r, z],
            [x - r, y + r, z],
            [x - r, y - r, z],
        ]]

    def ellipseType(self, element, geom, prop):
        return self.rectangleType(element, geom, prop)

    def pointType(self, element, geom, prop):
        geom['type'] = 'Point'
        geom['coordinates'] = element['center']

    def polylineType(self, element, geom, prop):
        if element['closed']:
            geom['type'] = 'Polygon'
            geom['coordinates'] = [element['points'][:]]
            geom['coordinates'][0].append(geom['coordinates'][0][0])
            if element.get('holes'):
                for hole in element['holes']:
                    hole = hole[:]
                    hole.append(hole[0])
                    geom['coordinates'].append(hole)
        else:
            geom['type'] = 'LineString'
            geom['coordinates'] = element['points']

    def rectangleType(self, element, geom, prop):
        x, y, z = element['center']
        width = element['width']
        height = element['height']
        rotation = element.get('rotation', 0)
        left = x - width / 2
        right = x + width / 2
        top = y - height / 2
        bottom = y + height / 2

        geom['type'] = 'Polygon'
        geom['coordinates'] = [[
            self.rotate(rotation, x, y, left, top, z),
            self.rotate(rotation, x, y, right, top, z),
            self.rotate(rotation, x, y, right, bottom, z),
            self.rotate(rotation, x, y, left, bottom, z),
            self.rotate(rotation, x, y, left, top, z),
        ]]

    # not represented
    # heatmap, griddata, image, pixelmap, arrow, rectanglegrid
    # heatmap could be MultiPoint, griddata could be rectangle with lots of
    # properties, image and pixelmap could be rectangle with the image id as a
    # property, arrow and rectangelgrid aren't really supported

    def elementToGeoJSON(self, element):
        elemType = element.get('type', '')
        funcName = elemType + 'Type'
        if not hasattr(self, funcName):
            return None
        result = {
            'type': 'Feature',
            'geometry': {},
            'properties': {
                k: v if k != 'id' else str(v)
                for k, v in element.items() if k in {
                    'id', 'label', 'group', 'user', 'lineColor', 'lineWidth',
                    'fillColor', 'radius', 'width', 'height', 'rotation',
                    'normal',
                }
            },
        }
        getattr(self, funcName)(element, result['geometry'], result['properties'])
        if result['geometry']['type'].lower() != element['type']:
            result['properties']['type'] = element['type']
        return result

    @property
    def geojson(self):
        return ''.join(self)


class GeoJSONAnnotation:
    def __init__(self, geojson):
        if not isinstance(geojson, (dict, list, tuple)):
            geojson = json.loads(geojson)
        self._elements = []
        self._annotation = {'elements': self._elements}
        self._parseFeature(geojson)

    def _parseFeature(self, geoelem):
        if isinstance(geoelem, (list, tuple)):
            for entry in geoelem:
                self._parseFeature(entry)
        if not isinstance(geoelem, dict) or 'type' not in geoelem:
            return
        if geoelem['type'] == 'FeatureCollection':
            return self._parseFeature(geoelem.get('features', []))
        if geoelem['type'] == 'GeometryCollection' and isinstance(geoelem.get('geometries'), list):
            for entry in geoelem['geometry']:
                self._parseFeature({'type': 'Feature', 'geometry': entry})
            return
        if geoelem['type'] in {'Point', 'LineString', 'Polygon', 'MultiPoint',
                               'MultiLineString', 'MultiPolygon'}:
            geoelem = {'type': 'Feature', 'geometry': geoelem}
        element = {k: v for k, v in geoelem.get('properties', {}).items() if k in {
            'id', 'label', 'group', 'user', 'lineColor', 'lineWidth',
            'fillColor', 'radius', 'width', 'height', 'rotation',
            'normal',
        }}
        if 'annotation' in geoelem.get('properties', {}):
            self._annotation.update(geoelem['properties']['annotation'])
            self._annotation['elements'] = self._elements
        elemtype = geoelem.get('properties', {}).get('type', '') or geoelem['geometry']['type']
        func = getattr(self, elemtype.lower() + 'Type', None)
        if func is not None:
            result = func(geoelem['geometry'], element)
            if isinstance(result, list):
                self._elements.extend(result)
            else:
                self._elements.append(result)

    def circleType(self, elem, result):
        cx = sum(e[0] for e in elem['coordinates'][0][:4]) / 4
        cy = sum(e[1] for e in elem['coordinates'][0][:4]) / 4
        try:
            cz = elem['coordinates'][0][0][2]
        except Exception:
            cz = 0
        radius = (max(e[0] for e in elem['coordinates'][0][:4]) -
                  min(e[0] for e in elem['coordinates'][0][:4])) / 2
        result['type'] = 'circle'
        result['radius'] = radius
        result['center'] = [cx, cy, cz]
        return result

    def ellipseType(self, elem, result):
        result = self.rectangleType(elem, result)
        result['type'] = 'ellipse'
        return result

    def rectangleType(self, elem, result):
        coor = elem['coordinates'][0]
        cx = sum(e[0] for e in coor[:4]) / 4
        cy = sum(e[1] for e in coor[:4]) / 4
        try:
            cz = elem['coordinates'][0][0][2]
        except Exception:
            cz = 0
        width = ((coor[0][0] - coor[1][0]) ** 2 + (coor[0][1] - coor[1][1]) ** 2) ** 0.5
        height = ((coor[1][0] - coor[2][0]) ** 2 + (coor[1][1] - coor[2][1]) ** 2) ** 0.5
        rotation = math.atan2(coor[1][1] - coor[0][1], coor[1][0] - coor[0][0])
        result['center'] = [cx, cy, cz]
        result['width'] = width
        result['height'] = height
        result['rotation'] = rotation
        result['type'] = 'rectangle'
        return result

    def pointType(self, elem, result):
        result['center'] = (elem['coordinates'] + [0, 0, 0])[:3]
        result['type'] = 'point'
        return result

    def multipointType(self, elem, result):
        results = []
        result['type'] = 'point'
        for entry in elem['coordinates']:
            subresult = result.copy()
            subresult['center'] = (entry + [0, 0, 0])[:3]
            results.append(subresult)
        return results

    def polylineType(self, elem, result):
        if elem.get('type') == 'LineString':
            return self.linestringType(elem, result)
        return self.polygonType(elem, result)

    def polygonType(self, elem, result):
        result['points'] = [(pt + [0])[:3] for pt in elem['coordinates'][0][:-1]]
        if len(elem['coordinates']) > 1:
            result['holes'] = [
                [(pt + [0])[:3] for pt in loop[:-1]]
                for loop in elem['coordinates'][1:]
            ]
        result['closed'] = True
        result['type'] = 'polyline'
        return result

    def multipolygonType(self, elem, result):
        results = []
        result['closed'] = True
        result['type'] = 'polyline'
        for entry in elem['coordinates']:
            subresult = result.copy()
            subresult['points'] = [(pt + [0])[:3] for pt in entry[0][:-1]]
            if len(entry) > 1:
                subresult['holes'] = [
                    [(pt + [0])[:3] for pt in loop[:-1]]
                    for loop in entry[1:]
                ]
            results.append(subresult)
        return results

    def linestringType(self, elem, result):
        result['points'] = [(pt + [0])[:3] for pt in elem['coordinates']]
        result['closed'] = False
        result['type'] = 'polyline'
        return result

    def multilinestringType(self, elem, result):
        results = []
        result['closed'] = False
        result['type'] = 'polyline'
        for entry in elem['coordinates']:
            subresult = result.copy()
            subresult['points'] = [(pt + [0])[:3] for pt in entry]
            results.append(subresult)
        return results

    def annotationToJSON(self):
        return json.dumps(self._annotation)

    @property
    def annotation(self):
        return self._annotation

    @property
    def elements(self):
        return self._elements

    @property
    def elementCount(self):
        return len(self._elements)


def isGeoJSON(annotation):
    """
    Check if a list or dictionary appears to contain a GeoJSON record.

    :param annotation: a list or dictionary.
    :returns: True if this appears to be GeoJSON
    """
    if isinstance(annotation, list):
        if len(annotation) < 1:
            return False
        annotation = annotation[0]
    if not isinstance(annotation, dict) or 'type' not in annotation:
        return False
    return annotation['type'] in {
        'Feature', 'FeatureCollection', 'GeometryCollection', 'Point',
        'LineString', 'Polygon', 'MultiPoint', 'MultiLineString',
        'MultiPolygon'}


class PlottableItemData:
    maxItems = 1000
    maxAnnotationElements = 10000
    maxDistinct = 20
    allowedTypes = (str, bool, int, float)

    def __init__(self, user, item, annotations=None, adjacentItems=False):
        """
        Get plottable data associated with an item.

        :param user: authenticating user.
        :param item: the item record.
        :param annotations: None, a list of annotation ids, or __all__.  If
            adjacent items are included, the most recent annotation with the
            same name will also be included.
        :param adjacentItems: if True, include data other items in the same
            folder.
        """
        self.user = user
        self._columns = None
        self._datacolumns = None
        self._data = None
        self._findItems(item, adjacentItems)
        self._findAnnotations(annotations)

    def _findItems(self, item, adjacentItems=False):
        self._columns = None
        self.item = item
        self.folder = Folder().load(id=item['folderId'], user=self.user, level=AccessType.READ)
        self.items = [item]
        if adjacentItems:
            for entry in Folder().childItems(self.folder):
                if len(self.items) >= self.maxItems:
                    break
                if entry['_id'] != item['_id']:
                    # skip if item doesn't have appropriate metadata or
                    # annotations.  If skipping, add to list to check if
                    # dataframe
                    # TODO:
                    self.items.append(entry)
        # TODO: find csv/xlsx/dataframe items in the folder, exclude them from
        # the item list but include them in general

    def _findAnnotations(self, annotations):
        from ..models.annotation import Annotation

        self._columns = None
        if isinstance(annotations, str):
            annotations = annotations.split(',')
        self.annotations = None
        if annotations and len(annotations):
            self.annotations = []
            query = {'_active': {'$ne': False}, 'itemId': self.item['_id']}
            if annotations[0] != '__all__':
                query['_id'] = {'$in': [ObjectId(annotId) for annotId in annotations]}
            self.annotations.append(list(Annotation().find(
                query, limit=0, sort=[('_version', -1)])))
            if not len(self.annotations[0]):
                self.annotations = None
        # Find adjacent annotations
        if annotations and len(self.items) > 1:
            names = {}
            for idx, annot in enumerate(self.annotations[0]):
                if annot['annotation']['name'] not in names:
                    names[annot['annotation']['name']] = idx
            for adjitem in self.items[1:]:
                query = {'_active': {'$ne': False}, 'itemId': adjitem['_id']}
                annotList = [None] * len(self.annotations[0])
                for annot in Annotation().find(query, limit=0, sort=[('_version', -1)]):
                    if annot['annotation']['name'] in names and annotList[
                            names[annot['annotation']['name']]] is None:
                        annotList[names[annot['annotation']['name']]] = annot
                self.annotations.append(annotList)

    def _addColumn(self, columns, fullkey, title, root, key, source):
        # Root should probably only be part of this at folder/csv level
        distinct = [source] if source not in {'folder'} else [source, root]
        if fullkey not in columns:
            columns[fullkey] = {
                'key': fullkey,
                'type': 'number',
                'where': [[distinct, key]], 'title': title,
                'count': 0, 'distinct': set(), 'min': None,
                'max': None}
            return (tuple(distinct), 0)
        elif [root, key, source] not in columns[fullkey]['where']:
            columns[fullkey]['where'].append([distinct, key])
        where = -1
        for colwhere in columns[fullkey]['where']:
            if tuple(colwhere[0]) == tuple(distinct):
                where += 1
                if colwhere[1] == key:
                    return (tuple(distinct), where)
        return (tuple(distinct), where)

    def _columnKey(self, source, root, key):
        if not hasattr(self, '_columnKeyCache'):
            self._columnKeyCache = {}
        hashkey = (source, root, key)
        if hashkey in self._columnKeyCache:
            return self._columnKeyCache[hashkey]
        fullkey = f'{root}.{key}.{source}'.lower()
        title = f'{root} {key}' if root is not None and root != '' else f'{key}'
        keymap = {
            r'(?i)(item|image)_(id|name)$': {'key': '_0_item.name', 'title': 'Item Name'},
            r'(?i)(low|min)(_|)x': {'key': '_bbox.x0', 'title': 'Bounding Box Low X'},
            r'(?i)(low|min)(_|)y': {'key': '_bbox.y0', 'title': 'Bounding Box Low Y'},
            r'(?i)(high|max)(_|)x': {'key': '_bbox.x1', 'title': 'Bounding Box High X'},
            r'(?i)(high|max)(_|)y': {'key': '_bbox.y1', 'title': 'Bounding Box High Y'},
        }
        for k, v in keymap.items():
            if re.match(k, key):
                fullkey = v['key']
                title = v['title']
                break
        self._columnKeyCache[hashkey] = fullkey, title
        return fullkey, title

    def _scanColumnByKey(self, result, key, entry, where=0, auxidx=0,
                         auxidx2=0, item=None, annotation=None):
        if result['type'] == 'number':
            try:
                [float(record[key]) for record in entry
                 if isinstance(record.get(key), self.allowedTypes)]
            except Exception:
                result['type'] = 'string'
                result['distinct'] = {str(v) for v in result['distinct']}
        for ridx, record in enumerate(entry):
            v = record.get(key)
            if not isinstance(v, self.allowedTypes):
                continue
            result['count'] += 1
            v = float(v) if result['type'] == 'number' else str(v)
            if len(result['distinct']) <= self.maxDistinct:
                result['distinct'].add(v)
            if result['type'] == 'number':
                if result['min'] is None:
                    result['min'] = result['max'] = v
                result['min'] = min(result['min'], v)
                result['max'] = max(result['max'], v)
            if self._datacolumns and result['key'] in self._datacolumns:
                self._datacolumns[result['key']][(where, (auxidx, auxidx2, ridx))] = v
                if item is not None:
                    self._datacolumns.get('_0_item.name', {})[
                        (where, (auxidx, auxidx2, ridx))] = item['name']
                    self._datacolumns.get('_2_item.id', {})[
                        (where, (auxidx, auxidx2, ridx))] = str(item['_id'])
                if annotation is not None:
                    self._datacolumns.get('_1_annotation.name', {})[
                        (where, (auxidx, auxidx2, ridx))] = annotation.get(
                            'annotation', {}).get('name')
                    self._datacolumns.get('_3_annotation.id', {})[
                        (where, (auxidx, auxidx2, ridx))] = str(annotation['_id'])
                    self._datacolumns.get('_4_annotation.description', {})[
                        (where, (auxidx, auxidx2, ridx))] = annotation.get(
                            'annotation', {}).get('description')

    def _scanColumn(self, meta, source, columns, auxmeta=None, auxidx2=0,
                    items=None, annotations=None):
        for root, entry in (list(meta.items()) + [(None, [meta])]):
            if not isinstance(entry, list) or not len(entry) or not isinstance(entry[0], dict):
                continue
            for key in entry[0]:
                if not isinstance(entry[0][key], self.allowedTypes):
                    continue
                fullkey, title = self._columnKey(source, root, key)
                where = self._addColumn(
                    columns, fullkey, title, root, key, source)
                result = columns[fullkey]
                self._scanColumnByKey(
                    result, key, entry, where, 0, auxidx2,
                    items[0] if items and len(items) > 0 else None,
                    annotations[0] if annotations and len(annotations) > 0 else None)
                if auxmeta:
                    for auxidx, aux in enumerate(auxmeta):
                        if root is None:
                            if isinstance(aux, dict) and key in aux:
                                self._scanColumnByKey(
                                    result, key, [aux], where, auxidx + 1, auxidx2,
                                    items[auxidx + 1] if items and len(items) > auxidx + 1 else
                                    None,
                                    annotations[auxidx + 1]
                                    if annotations and len(annotations) > auxidx + 1 else
                                    None)
                        elif (isinstance(aux.get(root), list) and
                                len(aux[root]) and
                                isinstance(aux[root][0], dict) and
                                key in aux[root][0]):
                            self._scanColumnByKey(
                                result, key, aux[root], where, auxidx + 1, auxidx2,
                                items[auxidx + 1] if items and len(items) > auxidx + 1 else None,
                                annotations[auxidx + 1] if annotations and
                                len(annotations) > auxidx + 1 else None)

    def _scanElementColumns(self, source, columns, elems, auxidx, items, annotations, keys):
        rows = {}
        for ridx, elem in enumerate(elems):
            if auxidx < len(items) and items[auxidx]:
                rows.setdefault('_0_item.name', []).append(items[auxidx]['name'])
                rows.setdefault('_2_item.id', []).append(str(items[auxidx]['_id']))
            if auxidx < len(annotations) and annotations[auxidx]:
                rows.setdefault('_1_annotation.name', []).append(
                    annotations[auxidx].get('annotation', {}).get('name'))
                rows.setdefault('_3_annotation.id', []).append(str(annotations[auxidx]['_id']))
                rows.setdefault('_4_annotation.description', []).append(
                    annotations[auxidx].get('annotation', {}).get('description'))
            if '_bbox' in elem:
                rows.setdefault('_bbox.x0', []).append(elem['_bbox']['lowx'])
                rows.setdefault('_bbox.y0', []).append(elem['_bbox']['lowy'])
                rows.setdefault('_bbox.x1', []).append(elem['_bbox']['highx'])
                rows.setdefault('_bbox.y1', []).append(elem['_bbox']['highy'])
            # TODO: Add group and label
            if not auxidx and not ridx and 'user' in elem:
                for key, entry in elem['user'].items():
                    if not isinstance(entry, self.allowedTypes):
                        continue
                    root = ''
                    fullkey, title = self._columnKey(source, root, key)
                    colwhere = self._addColumn(
                        columns, fullkey, title, root, key, source)
                    keys[key] = (fullkey, colwhere)
            # TODO: Populate group and label
            for key, (fullkey, _keywhere) in keys.items():
                entry = elem.get('user', {}).get(key)
                if not isinstance(entry, self.allowedTypes):
                    entry = None
                rows.setdefault(fullkey, []).append(entry)
        return rows

    def _scanElements(self, elements, source, columns, auxidx2, items, annotations):
        where = self._addColumn(
            columns, '_0_item.name', 'Item Name', '', 'name', 'annotationelement')
        self._addColumn(
            columns, '_2_item.id', 'Item ID', '', '_id', 'annotationelement')
        self._addColumn(
            columns, '_1_annotation.name', 'Annotation Name', '', 'name', 'annotationelement')
        self._addColumn(
            columns, '_3_annotation.id', 'Annotation ID', '', '_id', 'annotationelement')
        self._addColumn(
            columns, '_4_annotation.description', 'Annotation Description', '',
            'description', 'annotationelement')
        self._addColumn(
            columns, '_bbox.x0', 'Bounding Box Low X', '', 'lowx', 'annotationelement')
        self._addColumn(
            columns, '_bbox.y0', 'Bounding Box Low Y', '', 'lowy', 'annotationelement')
        self._addColumn(
            columns, '_bbox.x1', 'Bounding Box High X', '', 'highx', 'annotationelement')
        self._addColumn(
            columns, '_bbox.y1', 'Bounding Box High Y', '', 'highy', 'annotationelement')
        keys = {}
        for auxidx, elems in enumerate(elements):
            if not elems:
                continue
            rows = self._scanElementColumns(
                source, columns, elems, auxidx, items, annotations, keys)
            for fullkey, entry in rows.items():
                result = columns[fullkey]
                if result['type'] == 'number':
                    try:
                        [float(v) for v in entry if isinstance(v, self.allowedTypes)]
                    except Exception:
                        result['type'] = 'string'
                        result['distinct'] = {str(v) for v in result['distinct']}
                for ridx, v in enumerate(entry):
                    if not isinstance(v, self.allowedTypes):
                        continue
                    result['count'] += 1
                    v = float(v) if result['type'] == 'number' else str(v)
                    if len(result['distinct']) <= self.maxDistinct:
                        result['distinct'].add(v)
                    if result['type'] == 'number':
                        if result['min'] is None:
                            result['min'] = result['max'] = v
                        result['min'] = min(result['min'], v)
                        result['max'] = max(result['max'], v)
                    if self._datacolumns and fullkey in self._datacolumns:
                        self._datacolumns[fullkey][(where, (auxidx, auxidx2, ridx))] = v

    @property
    def columns(self):
        """
        Get a sorted list of plottable columns with some metadata for each.

        Each data entry contains

            :fullkey: a unique string.  This is a good first-order sort
            :root: the root data array
            :key: the specific data tag
            :source: the source of the data (folder, item, annotation,
                annotationelement, file)
            :type: string or number
            :title: a human readable title
            :[distinct]: a list of distinct values if there are less than some
                maximum number of distinct values.  This might not include i
                values from adjacent items
            :[min]: for number data types, the lowest value present
            :[max]: for number data types, the highest value present

        :returns: a sorted list of data entries.
        """
        from ..models.annotationelement import Annotationelement

        if self._columns is not None:
            return self._columns
        columns = {}
        self._addColumn(
            columns, '_0_item.name', 'Item Name', 'Item', 'name', 'base')
        self._addColumn(
            columns, '_2_item.id', 'Item ID', 'Item', '_id', 'base')
        self._scanColumn(self.folder.get('meta', {}), 'folder', columns)
        self._scanColumn(self.item.get('meta', {}), 'item', columns,
                         [item.get('meta', {}) for item in self.items[1:]],
                         items=self.items)
        for anidx, annot in enumerate(self.annotations[0] if self.annotations is not None else []):
            self._scanColumn(
                annot.get('annotation', {}).get('attributes', {}),
                'annotation', columns,
                [itemannot[anidx].get('annotation').get('attributes', {})
                 for itemannot in self.annotations[1:]
                 if itemannot[anidx] is not None],
                anidx, items=self.items,
                annotations=[a[anidx] for a in self.annotations])
            if not anidx:
                self._addColumn(
                    columns, '_1_annotation.name', 'Annotation Name',
                    'Annotation', 'name', 'base')
                self._addColumn(
                    columns, '_3_annotation.id', 'Annotation ID',
                    'Annotation', '_id', 'base')
                self._addColumn(
                    columns, '_4_annotation.description', 'Annotation Description',
                    'Annotation', 'description', 'base')
            # add annotation elements
            firstelem = next(Annotationelement().yieldElements(annot), None)
            if firstelem is not None:
                self._scanElements(
                    [list(itertools.islice(
                        Annotationelement().yieldElements(a[anidx], bbox=True), 10000))
                        if a[anidx] else None for a in self.annotations],
                    'annotationelement', columns,
                    anidx,
                    items=self.items,
                    annotations=[a[anidx] for a in self.annotations],
                )
        # TODO: Add csv
        for result in columns.values():
            if len(result['distinct']) <= self.maxDistinct:
                result['distinct'] = sorted(result['distinct'])
                result['distinctcount'] = len(result['distinct'])
            else:
                result.pop('distinct', None)
            if result['type'] != 'number' or result['min'] is None:
                result.pop('min', None)
                result.pop('max', None)
        self._columns = sorted(columns.values(), key=lambda x: x['key'])
        return self._columns

    def data(self, columns, requiredColumns=None):
        """
        Get plottable data.

        :param columns: the columns to return.  Either a list of column names
            or a comma-delimited string.
        :param requiredColumns: only return data rows where all of these
            columns are non-None.  Either a list of column names of a
            comma-delimited string.
        """
        if not isinstance(columns, list):
            columns = columns.split(',')
        if not isinstance(requiredColumns, list):
            requiredColumns = requiredColumns.split(',') if requiredColumns is not None else []
        self._datacolumns = {c: {} for c in columns}
        rows = set()
        # collects data as a side effect
        collist = self.columns
        for coldata in self._datacolumns.values():
            rows |= set(coldata.keys())
        rows = sorted(rows, key=lambda row: (
            tuple(x if x is not None else '' for x in row[0]), row[1]))
        colsout = [col.copy() for col in collist if col['key'] in columns]
        for cidx, col in enumerate(colsout):
            col['index'] = cidx
        logger.info(f'Gathering {len(self._datacolumns)} x {len(rows)} data')
        data = [[None] * len(self._datacolumns) for _ in range(len(rows))]
        for cidx, col in enumerate(colsout):
            colkey = col['key']
            if colkey in self._datacolumns:
                datacol = self._datacolumns[colkey]
                for ridx, rowid in enumerate(rows):
                    data[ridx][cidx] = datacol.get(rowid, None)
        for cidx, col in enumerate(colsout):
            colkey = col['key']
            numrows = len(data)
            if colkey in requiredColumns:
                data = [row for row in data if row[cidx] is not None]
            if len(data) < numrows:
                logger.info(f'Reduced row count from {numrows} to {len(data)} '
                            f'because of None values in column {colkey}')
        # Refresh our count, distinct, distinctcount, min, max for each column
        for cidx, col in enumerate(colsout):
            col['count'] = len([row[cidx] for row in data if row[cidx] is not None])
            if col['type'] == 'number' and col['count']:
                col['min'] = min(row[cidx] for row in data if row[cidx] is not None)
                col['max'] = max(row[cidx] for row in data if row[cidx] is not None)
            distinct = {str(row[cidx]) for row in data if row[cidx] is not None}
            if len(distinct) <= self.maxDistinct:
                col['distinct'] = sorted(distinct)
                col['distinctcount'] = len(distinct)
            else:
                col.pop('distinct', None)
                col.pop('distinctcount', None)
        return {
            'columns': colsout,
            'data': data}
